import json
import logging
from datetime import datetime

import httpx
import pytest

from conftest import get_mock_filename, get_mock_response
from connectsensor import (
    APIVersion,
    KingspanAPIError,
    KingspanInvalidCredentialsError,
    KingspanTimeoutError,
    SensorClient,
)
from mock_data import PASSWORD, USERNAME


def test_status(mock_sync_httpx_post, mock_wsdl):  # noqa: ARG001
    def check_client(api_version: APIVersion) -> None:
        client = SensorClient(api_version)
        client.login(USERNAME, PASSWORD)
        tanks = client.tanks
        assert tanks[0].level == 1000
        assert tanks[0].serial_number == "20001000"
        assert tanks[0].model == "TestModel"
        assert tanks[0].name == "TestTank"
        assert tanks[0].capacity == 2000
        assert tanks[0].last_read.replace(microsecond=0) == datetime(
            2021,
            1,
            31,
            0,
            59,
            30,
        )
        tank_history = tanks[0].history()
        assert tank_history[0]["reading_date"] == datetime(2021, 1, 25, 13, 59, 14)
        assert tank_history[1]["level_percent"] == 95
        assert tank_history[2]["level_litres"] == 1880

    check_client(APIVersion.CONNECT_V1)
    check_client(APIVersion.KNECT_V1)


def test_login_exception(mock_sync_httpx_post):  # noqa: ARG001
    client = SensorClient()
    with pytest.raises(
        KingspanInvalidCredentialsError,
        match="Authentication Failed, Invalid Login",
    ):
        client.login("invalid_user", "invalid_password")


def test_unknown_api_error(mocker):
    mock_response = httpx.Response(
        status_code=200,
        content=json.dumps(
            {"apiResult": {"code": 1, "description": "test API error"}}
        ).encode("utf-8"),
    )
    mocker.patch.object(httpx.Client, "post", return_value=mock_response)

    client = SensorClient()
    with pytest.raises(
        KingspanAPIError,
        match="test API error",
    ):
        client.login(USERNAME, PASSWORD)


def test_tank_exception(mocker):
    def mocked_post(self, url, *args, **kwargs):  # noqa: ANN003, ARG001
        if "GetLatestLevel" in url:
            raise KingspanAPIError("Test Exception for GetLatestLevel")
        mock_filename = get_mock_filename(url, kwargs["content"])
        return get_mock_response(url, open(mock_filename, "rb").read())

    mocker.patch.object(httpx.Client, "post", new=mocked_post)

    client = SensorClient()
    client.login(USERNAME, PASSWORD)
    tanks = client.tanks

    with pytest.raises(
        KingspanAPIError,
        match="Test Exception for GetLatestLevel",
    ):
        _ = tanks[0].level


def test_timeout(mocker):
    mocker.patch.object(
        httpx.Client,
        "post",
        side_effect=httpx.TimeoutException("Test httpx timeout"),
    )

    client = SensorClient()

    with pytest.raises(
        KingspanTimeoutError,
        match="HTTP request timeout: Test httpx timeout",
    ):
        client.login(USERNAME, PASSWORD)


def test_api_unauthorized(mocker):
    mock_response = httpx.Response(status_code=401, content=b"Unauthorized")

    mocker.patch.object(httpx.Client, "post", return_value=mock_response)

    client = SensorClient()
    with pytest.raises(KingspanAPIError, match="Invalid token authorization"):
        client.login(USERNAME, PASSWORD)


def test_malformed_response(mocker):
    mock_response = httpx.Response(
        status_code=200, content=json.dumps({"unexpected": "data"}).encode("utf-8")
    )

    mocker.patch.object(httpx.Client, "post", return_value=mock_response)

    client = SensorClient()
    with pytest.raises(
        KingspanAPIError,
        match="Malformed response from API: cannot extract response/code",
    ):
        client.login(USERNAME, PASSWORD)


def test_history_exception(mocker):
    def mocked_post(self, url, *args, **kwargs):
        if "GetCallHistory" in url:
            msg = "Test Exception for GetCallHistory"
            raise KingspanAPIError(msg)
        mock_filename = get_mock_filename(url, kwargs["content"])
        return get_mock_response(url, open(mock_filename, "rb").read())

    mocker.patch.object(httpx.Client, "post", new=mocked_post)
    client = SensorClient()
    client.login(USERNAME, PASSWORD)
    tanks = client.tanks

    with pytest.raises(
        KingspanAPIError,
        match="Test Exception for GetCallHistory",
    ):
        _ = tanks[0].history()


def test_debug_redaction(mock_sync_httpx_post, caplog):  # noqa: ARG001
    logger = logging.getLogger("connectsensor")
    logger.propagate = True
    caplog.set_level(logging.DEBUG, logger="connectsensor")
    caplog.clear()

    client = SensorClient()
    client.login(USERNAME, PASSWORD)
    log_text = caplog.text
    assert len(log_text.splitlines()) == 2
    assert USERNAME not in log_text
    assert PASSWORD not in log_text
    assert "*redacted*" in log_text



"""Constants and types shared across the package."""

from datetime import datetime
from enum import IntEnum, auto
from typing import TypeAlias

# JSON API is HTTP and sends the username and password in the clear
API_SERVER = "sensorapi.connectsensor.com"
API_PORT = 8087
API_BASE_URL = f"http://{API_SERVER}:{API_PORT}"

# SOAP API
DEFAULT_SERVER = "www.connectsensor.com"
WSDL_PATH = "soap/MobileApp.asmx?WSDL"
WSDL_URL = f"https://{DEFAULT_SERVER}/{WSDL_PATH}"
SOAP_NS_PREFIX = "http://mobileapp/"


# API versions don't match what Kingspan call them as these are not obviously
# incrementing and use different servers/protocols.
class APIVersion(IntEnum):
    CONNECT_V1 = auto()
    KNECT_V1 = auto()


DEFAULT_API_VERSION = APIVersion.KNECT_V1

HTTP_UNAUTHORIZED = 401

# The JWT appears to be hard-coded into the application rather than securely returned
TOKEN = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJUaGVNb2JpbGVBcHAiLCJyb2xlIjoiVGhlTW9iaWxlQXBwIiwiZXhwIjoxNzg2ODk4NTM3LCJpc3MiOiJTZW5zb3JBUEkgQXV0aFNlcnZlciIsImF1ZCI6IlNlbnNvckFQSSBVc2VycyJ9.PW-NP46vP9pP5Da87KIzsN6ZWIA3vOI4XbqxHWVuTOY"  # noqa: E501, S105, cspell:disable-line

# Types for API request and response
APIResponseValue: TypeAlias = (
    str | int | datetime | list["APIResponseValue"] | dict[str, "APIResponseValue"]
)
APIRequest: TypeAlias = dict[str, APIResponseValue]
APIResponse: TypeAlias = dict[str, APIResponseValue]



"""Kingspan Sensit API."""

import importlib.metadata

from connectsensor.client import AsyncSensorClient, SensorClient  # noqa: F401
from connectsensor.const import APIVersion
from connectsensor.exceptions import *  # noqa: F403

__version__ = importlib.metadata.version("kingspan-connect-sensor")



from _typeshed import Incomplete
from connectsensor.client import (
    AsyncSensorClient as AsyncSensorClient,
    SensorClient as SensorClient,
)
from connectsensor.exceptions import (
    KingspanAPIError as KingspanAPIError,
    KingspanDBError as KingspanDBError,
    KingspanInvalidCredentialsError as KingspanInvalidCredentialsError,
)

__version__: Incomplete



from connectsensor.exceptions import *
from _typeshed import Incomplete
from connectsensor.client import AsyncSensorClient as AsyncSensorClient, SensorClient as SensorClient
from connectsensor.const import APIVersion as APIVersion

__version__: Incomplete



from datetime import datetime

from async_property import async_property

from connectsensor.const import APIResponse

class _BaseTank:
    def __init__(self, client: object, signalman_no: str) -> None: ...
    def unpack_tank_data(self, response: APIResponse) -> None: ...
    def transform_history_data(self, data: dict[str, list[dict]]) -> list[dict]: ...

class Tank(_BaseTank):
    @property
    def level(self) -> int: ...
    @property
    def serial_number(self) -> str: ...
    @property
    def model(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def capacity(self) -> int: ...
    @property
    def last_read(self) -> datetime: ...
    def history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> APIResponse: ...

class AsyncTank(_BaseTank):
    @async_property
    async def level(self) -> int: ...
    @async_property
    async def serial_number(self) -> str: ...
    @async_property
    async def model(self) -> str: ...
    @async_property
    async def name(self) -> str: ...
    @async_property
    async def capacity(self) -> int: ...
    @async_property
    async def last_read(self) -> datetime: ...
    async def history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> APIResponse: ...



from async_property import async_property
from connectsensor.const import APIResponse as APIResponse
from datetime import datetime

class _BaseTank:
    def __init__(self, client: object, signalman_no: str) -> None: ...
    def unpack_tank_data(self, response: APIResponse) -> None: ...
    def format_reading_date(self, date: datetime | str) -> datetime: ...
    def transform_history_data(self, data: APIResponse) -> APIResponse: ...

class Tank(_BaseTank):
    @property
    def level(self) -> int: ...
    @property
    def serial_number(self) -> str: ...
    @property
    def model(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def capacity(self) -> int: ...
    @property
    def last_read(self) -> datetime: ...
    @property
    def history(self) -> APIResponse: ...

class AsyncTank(_BaseTank):
    @async_property
    async def level(self) -> int: ...
    @async_property
    async def serial_number(self) -> str: ...
    @async_property
    async def model(self) -> str: ...
    @async_property
    async def name(self) -> str: ...
    @async_property
    async def capacity(self) -> int: ...
    @async_property
    async def last_read(self) -> datetime: ...
    @async_property
    async def history(self) -> APIResponse: ...



from _typeshed import Incomplete
from enum import IntEnum
from typing import TypeAlias

API_SERVER: str
API_PORT: int
API_BASE_URL: Incomplete
DEFAULT_SERVER: str
WSDL_PATH: str
WSDL_URL: Incomplete
SOAP_NS_PREFIX: str

class APIVersion(IntEnum):
    CONNECT_V1 = ...
    KNECT_V1 = ...

DEFAULT_API_VERSION: Incomplete
HTTP_UNAUTHORIZED: int
TOKEN: str
APIResponseValue: TypeAlias
APIRequest: TypeAlias = dict[str, APIResponseValue]
APIResponse: TypeAlias = dict[str, APIResponseValue]



from datetime import datetime
from typing import TypeAlias

from _typeshed import Incomplete

API_SERVER: str
API_PORT: int
API_BASE_URL: Incomplete
HTTP_UNAUTHORIZED: int
TOKEN: str

APIResponseValue: TypeAlias = (
    str | int | datetime | list[APIResponseValue] | dict[str, APIResponseValue]
)
APIRequest: TypeAlias = dict[str, APIResponseValue]
APIResponse: TypeAlias = dict[str, APIResponseValue]



import argparse
import logging
import sys

from connectsensor import (
    APIVersion,
    KingspanAPIError,
    KingspanInvalidCredentialsError,
    SensorClient,
)


class PasswordFilter(logging.Filter):
    def filter(self, record):
        if isinstance(record.msg, str):
            record.msg = record.msg.replace("password", "secret")
        return True


def enable_debug_logging() -> None:
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
    handler.setFormatter(formatter)
    handler.addFilter(PasswordFilter())
    logger = logging.getLogger("connectsensor")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    logger.propagate = False


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--username",
        required=True,
        help="Kingspan account email address",
    )
    parser.add_argument("--password", required=True, help="Kingspan account password")
    parser.add_argument(
        "--api-version",
        choices=APIVersion.__members__.keys(),
        default="KNECT_V1",
        help="Which API to use from Kingspan (default: KNECT_V1)",
    )
    parser.add_argument("--debug", action="store_true")

    args = parser.parse_args()

    if args.debug:
        enable_debug_logging()

    client = SensorClient(APIVersion[args.api_version])
    try:
        client.login(args.username, args.password)
    except KingspanInvalidCredentialsError:
        print("Authentication Failed: invalid username or password", file=sys.stderr)
        sys.exit(1)
    except KingspanAPIError as e:
        print(f"Kingspan API error: {e}", file=sys.stderr)
        sys.exit(1)

    for tank in client.tanks:
        tank_level_percent = str(int(100 * tank.level / tank.capacity))

        print(tank.name + ":")
        print(f"\tCapacity = {tank.capacity}")
        print(f"\tSerial Number = {tank.serial_number}")
        print(f"\tModel = {tank.model}")
        print(f"\tLevel = {tank_level_percent}% ({tank.level} litres)")
        print(f"\tLast Read = {tank.last_read}")

        print("\nHistory:")
        print("\t{:<22} {:<6} {:<5}".format("Reading date", "%Full", "Litres"))
        for _, measurement in enumerate(tank.history()):
            print(
                "\t{:<22} {:<6} {:<5}".format(
                    measurement["reading_date"].strftime("%d-%b-%Y %H:%M"),
                    measurement["level_percent"],
                    measurement["level_litres"],
                ),
            )


if __name__ == "__main__":
    # execute only if run as a script
    main()



from datetime import datetime

import httpx
from async_property import async_property

from connectsensor.const import (
    APIRequest,
    APIResponse,
    APIVersion,
)
from connectsensor.tank import AsyncTank, Tank

class _BaseClient:
    def __init__(self, version: APIVersion = ...) -> None: ...
    def redact(self, obj: APIResponse) -> APIResponse: ...
    def build_request(
        self,
        endpoint: str,
        data: dict[str, str | list | dict],
    ) -> tuple: ...
    def check_payload(self, response: httpx.Response) -> APIResponse: ...
    def httpx_exception(self, exc: Exception) -> None: ...
    def get_history_request(
        self,
        signalman_no: str,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> tuple[str, APIRequest]: ...

class SensorClient(_BaseClient):
    def __init__(self, version) -> None: ...
    def login(self, username: str, password: str) -> None: ...
    @property
    def tanks(self) -> list[Tank]: ...

class AsyncSensorClient(_BaseClient):
    async def __aenter__(self) -> AsyncSensorClient: ...
    async def __aexit__(self, exc_t, exc_v, exc_tb) -> None: ...
    async def login(self, username: str, password: str) -> None: ...
    @async_property
    async def tanks(self) -> list[AsyncTank]: ...



import httpx
from _typeshed import Incomplete
from collections.abc import Callable as Callable
from connectsensor.const import APIResponse as APIResponse, APIVersion as APIVersion, API_BASE_URL as API_BASE_URL, DEFAULT_API_VERSION as DEFAULT_API_VERSION, HTTP_UNAUTHORIZED as HTTP_UNAUTHORIZED, SOAP_NS_PREFIX as SOAP_NS_PREFIX, TOKEN as TOKEN, WSDL_URL as WSDL_URL
from connectsensor.exceptions import KingspanAPIError as KingspanAPIError, KingspanInvalidCredentialsError as KingspanInvalidCredentialsError, KingspanTimeoutError as KingspanTimeoutError
from connectsensor.tank import AsyncTank as AsyncTank, Tank as Tank
from datetime import datetime
from zeep.transports import Transport

POST_HEADERS: Incomplete

class HttpxTransport(Transport):
    client: Incomplete
    def __init__(self) -> None: ...
    def post(self, address: str, envelope: bytes | None, headers: dict) -> httpx.Response: ...

class _BaseClient:
    def __init__(self, version: APIVersion = ...) -> None: ...
    def redact_response(self, obj: APIResponse) -> APIResponse: ...
    def check_response(self, response: object) -> APIResponse: ...
    def handle_exception(self, exc: Exception) -> None: ...
    def login_request(self, username: str, password: str) -> tuple[Callable, list, dict]: ...
    def login_response(self, response: httpx.Response | object, tank_type: type) -> None: ...
    def get_latest_level_request(self, signalman_no: str) -> tuple[Callable, list, dict]: ...
    def get_history_request(self, signalman_no: str, start_date: datetime | None = None, end_date: datetime | None = None) -> tuple[Callable, list, dict]: ...

class SensorClient(_BaseClient):
    def __init__(self, version: APIVersion = ...) -> None: ...
    def login(self, username: str, password: str) -> None: ...
    @property
    def tanks(self) -> list[Tank]: ...

class AsyncSensorClient(_BaseClient):
    async def __aenter__(self) -> AsyncSensorClient: ...
    async def __aexit__(self, exc_t, exc_v, exc_tb) -> None: ...
    async def login(self, username: str, password: str) -> None: ...
    @property
    async def tanks(self) -> list[AsyncTank]: ...



from pytest import mark

from mock_data import VALID_STATUS


@mark.script_launch_mode("subprocess")
def test_help(script_runner):
    ret = script_runner.run("kingspan-status", print_result=False)
    assert not ret.success
    assert "usage: kingspan-status [-h]" in ret.stderr
    assert ret.stdout == ""


@mark.script_launch_mode("subprocess")
def test_help_verbose(script_runner):
    ret = script_runner.run("kingspan-status", "--help", print_result=False)
    assert ret.success
    assert "usage: kingspan-status [-h]" in ret.stdout
    assert "username USERNAME" in ret.stdout
    assert ret.stderr == ""


@mark.script_launch_mode("inprocess")
def test_help_invalid_credentials(script_runner, mock_sync_httpx_post):
    ret = script_runner.run(
        "kingspan-status",
        "--username=invalid@example.com",
        "--password=invalid",
        print_result=False,
    )

    assert not ret.success
    assert "invalid username or password" in ret.stderr
    assert ret.stdout == ""


@mark.script_launch_mode("inprocess")
def test_status(script_runner, mock_sync_httpx_post):
    ret = script_runner.run(
        "kingspan-status",
        "--username=test@example.com",
        "--password=s3cret",
        print_result=False,
    )
    assert ret.success
    output = [s.strip() for s in ret.stdout.split("\n")]
    assert output == VALID_STATUS
    assert ret.stderr == ""


@mark.script_launch_mode("inprocess")
def test_debug(script_runner, mock_sync_httpx_post, mock_wsdl):
    ret = script_runner.run(
        "kingspan-status",
        "--username=test@example.com",
        "--password=s3cret",
        "--debug",
        "--api-version=CONNECT_V1",
        print_result=False,
    )
    assert ret.success
    output = [s.strip() for s in ret.stdout.split("\n")]
    assert output == VALID_STATUS
    stderr = [s.strip() for s in ret.stderr.split("\n")]
    assert stderr[0] == "DEBUG:connectsensor:Init API with version=CONNECT_V1"
    assert "'apiUserID': '*redacted*'" in stderr[1]
    assert "'emailAddress': '*redacted*'" in stderr[1]



"""Tank class for Connect Sensor API."""

from datetime import datetime
from typing import cast

from async_property import async_property

from connectsensor.const import APIResponse, APIResponseValue


class _BaseTank:
    """Base class for Tanks."""

    def __init__(self, client: object, signalman_no: str) -> None:
        """Initialize a Tank instance."""
        self._client = client
        self._signalman_no = signalman_no
        self._level_data = None

    def unpack_tank_data(self, response: APIResponse) -> None:
        self._level_data = response["level"]
        self._tank_info = {x["name"]: x["value"] for x in response["tankInfo"]}

    def format_reading_date(self, date: datetime | str) -> datetime:
        if isinstance(date, datetime):
            return date
        return datetime.fromisoformat(date)

    def transform_history_data(
        self,
        data: APIResponse,
    ) -> list[dict[str, APIResponseValue]]:
        """Transform raw tank history data into a list of dicts."""
        result: list[dict[str, APIResponseValue]] = []

        for x in data:
            x_dict = cast("dict[str, APIResponseValue]", x)
            result.append(
                {
                    "reading_date": self.format_reading_date(x_dict["readingDate"]),
                    "level_percent": x_dict["levelPercentage"],
                    "level_litres": x_dict["levelLitres"],
                },
            )

        return result


class Tank(_BaseTank):
    """Synchronous Tank class."""

    @property
    def level(self) -> int:
        """Return the oil level in the tank in litres."""
        self._cache_tank_data()
        return int(self._level_data["levelLitres"])

    @property
    def serial_number(self) -> str:
        """Return the serial number of the tank."""
        self._cache_tank_data()
        return (
            self._tank_info["Signalman No"]
            if "Signalman No" in self._tank_info
            else self._tank_info["Serial No"]
        )

    @property
    def model(self) -> str:
        """Return the model of the tank."""
        self._cache_tank_data()
        return self._tank_info["Model"]

    @property
    def name(self) -> str:
        """Return the name of the tank."""
        self._cache_tank_data()
        return self._tank_info["Tank Name"]

    @property
    def capacity(self) -> int:
        """Return the capacity of the tank in litres."""
        self._cache_tank_data()
        return int(self._tank_info["Tank Capacity(L)"])

    @property
    def last_read(self) -> datetime:
        """Return the last read date of the tank as a datetime object."""
        self._cache_tank_data()
        return self.format_reading_date(self._level_data["readingDate"])

    def history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> APIResponse:
        """Return the history of the tank readings as a list of dicts."""
        history_data = self._client._get_history(  # noqa: SLF001
            self._signalman_no, start_date, end_date
        )
        return self.transform_history_data(history_data)

    def _cache_tank_data(self) -> None:
        """Cache the latest tank data if not already cached."""
        if self._level_data is None:
            data = self._client._get_latest_level(self._signalman_no)  # noqa: SLF001
            self.unpack_tank_data(data)


class AsyncTank(_BaseTank):
    """Asynchronous Tank class."""

    @async_property
    async def level(self) -> int:
        """Return the oil level in the tank in litres."""
        await self._cache_tank_data()
        return int(self._level_data["levelLitres"])

    @async_property
    async def serial_number(self) -> str:
        """Return the serial number of the tank."""
        await self._cache_tank_data()
        return (
            self._tank_info["Signalman No"]
            if "Signalman No" in self._tank_info
            else self._tank_info["Serial No"]
        )

    @async_property
    async def model(self) -> str:
        """Return the model of the tank."""
        await self._cache_tank_data()
        return self._tank_info["Model"]

    @async_property
    async def name(self) -> str:
        """Return the name of the tank."""
        await self._cache_tank_data()
        return self._tank_info["Tank Name"]

    @async_property
    async def capacity(self) -> int:
        """Return the capacity of the tank in litres."""
        await self._cache_tank_data()
        return int(self._tank_info["Tank Capacity(L)"])

    @async_property
    async def last_read(self) -> datetime:
        """Return the last read date of the tank as a datetime object."""
        await self._cache_tank_data()
        return self.format_reading_date(self._level_data["readingDate"])

    async def history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> APIResponse:
        """Return the history of the tank readings as a list of dicts."""
        history_data = await self._client._get_history(  # noqa: SLF001
            self._signalman_no, start_date, end_date
        )
        return self.transform_history_data(history_data)

    async def _cache_tank_data(self) -> None:
        """Cache the latest tank data if not already cached."""
        if self._level_data is None:
            response = await self._client._get_latest_level(  # noqa: SLF001
                self._signalman_no,
            )
            self.unpack_tank_data(response)



import logging
from connectsensor import APIVersion as APIVersion, KingspanAPIError as KingspanAPIError, KingspanInvalidCredentialsError as KingspanInvalidCredentialsError, SensorClient as SensorClient

class PasswordFilter(logging.Filter):
    def filter(self, record): ...

def enable_debug_logging() -> None: ...
def main() -> None: ...



import logging
from datetime import datetime

import httpx
import pytest

from conftest import get_mock_filename, get_mock_response
from connectsensor import (
    APIVersion,
    AsyncSensorClient,
    KingspanAPIError,
    KingspanInvalidCredentialsError,
)
from mock_data import PASSWORD, USERNAME


@pytest.mark.asyncio
async def test_status(mock_async_httpx_post, mock_wsdl):  # noqa: ARG001
    async def check_client(api_version: APIVersion) -> None:
        async with AsyncSensorClient(api_version) as client:
            await client.login(USERNAME, PASSWORD)
            tanks = await client.tanks
            assert await tanks[0].level == 1000
            assert await tanks[0].serial_number == "20001000"
            assert await tanks[0].model == "TestModel"
            assert await tanks[0].name == "TestTank"
            assert await tanks[0].capacity == 2000
            last_read = await tanks[0].last_read
            assert last_read.replace(microsecond=0) == datetime(2021, 1, 31, 0, 59, 30)
            tank_history = await tanks[0].history()
            assert tank_history[0]["reading_date"] == datetime(2021, 1, 25, 13, 59, 14)
            assert tank_history[1]["level_percent"] == 95
            assert tank_history[2]["level_litres"] == 1880

    await check_client(APIVersion.CONNECT_V1)
    await check_client(APIVersion.KNECT_V1)


@pytest.mark.asyncio
async def test_login_exception(mock_async_httpx_post):  # noqa: ARG001
    async with AsyncSensorClient() as client:
        with pytest.raises(
            KingspanInvalidCredentialsError,
            match="Authentication Failed, Invalid Login",
        ):
            await client.login("invalid_user", "invalid_password")


@pytest.mark.asyncio
async def test_tank_exception(mocker):
    async def mocked_post(self, url, *args, **kwargs) -> httpx.Response:
        if "GetLatestLevel" in url:
            msg = "Test Exception for GetLatestLevel"
            raise KingspanAPIError(msg)
        mock_filename = get_mock_filename(url, kwargs["content"])
        return get_mock_response(url, open(mock_filename, "rb").read())

    mocker.patch.object(httpx.AsyncClient, "post", new=mocked_post)

    async with AsyncSensorClient() as client:
        await client.login(USERNAME, PASSWORD)
        tanks = await client.tanks

        with pytest.raises(
            KingspanAPIError,
            match="Test Exception for GetLatestLevel",
        ):
            _ = await tanks[0].level


@pytest.mark.asyncio
async def test_history_exception(mocker):
    async def mocked_post(self, url, *args, **kwargs):
        if "GetCallHistory" in url:
            msg = "Test Exception for GetCallHistory"
            raise KingspanAPIError(msg)
        mock_filename = get_mock_filename(url, kwargs["content"])
        return get_mock_response(url, open(mock_filename, "rb").read())

    mocker.patch.object(httpx.AsyncClient, "post", new=mocked_post)

    async with AsyncSensorClient() as client:
        await client.login(USERNAME, PASSWORD)
        tanks = await client.tanks

        with pytest.raises(
            KingspanAPIError,
            match="Test Exception for GetCallHistory",
        ):
            _ = await tanks[0].history()


@pytest.mark.asyncio
async def test_debug_redaction(
    mock_async_httpx_post,  # noqa: ARG001
    mock_wsdl,  # noqa: ARG001
    caplog,
):
    async def check_client(api_version: APIVersion) -> None:
        logger = logging.getLogger("connectsensor")
        logger.propagate = True
        caplog.set_level(logging.DEBUG, logger="connectsensor")
        caplog.clear()
        async with AsyncSensorClient(api_version) as client:
            await client.login(USERNAME, PASSWORD)
            log_text = caplog.text
            assert len(log_text.splitlines()) == 2
            assert USERNAME not in log_text
            assert PASSWORD not in log_text
            assert "*redacted*" in log_text

    await check_client(APIVersion.CONNECT_V1)
    await check_client(APIVersion.KNECT_V1)



"""Exceptions possible in the Kingspan API."""


class KingspanDBError(Exception):
    """Exception raised for tank history database errors."""


class KingspanAPIError(Exception):
    """Base class for exceptions raised for internal errors from Kingspan API."""


class KingspanInvalidCredentialsError(KingspanAPIError):
    """Username and/or password were invalid."""


class KingspanTimeoutError(KingspanAPIError):
    """Kingspan API timed out."""



"""Client for interacting with the Kingspan Connect Sensor."""

import json
import logging
from asyncio import get_running_loop
from collections.abc import Callable
from datetime import datetime
from typing import cast

import httpx
from zeep import AsyncClient as AsyncSoapClient
from zeep import Client as SoapClient
from zeep.helpers import serialize_object
from zeep.transports import AsyncTransport, Transport

from connectsensor.const import (
    API_BASE_URL,
    DEFAULT_API_VERSION,
    HTTP_UNAUTHORIZED,
    SOAP_NS_PREFIX,
    TOKEN,
    WSDL_URL,
    APIResponse,
    APIVersion,
)
from connectsensor.exceptions import (
    KingspanAPIError,
    KingspanInvalidCredentialsError,
    KingspanTimeoutError,
)
from connectsensor.tank import AsyncTank, Tank

_LOGGER: logging.Logger = logging.getLogger(__package__)
_LOGGER.debug("Using Kingspan API: %s", API_BASE_URL)

POST_HEADERS = {
    "Content-Type": "application/json",
    "Accept": "application/json",
    "Authorization": f"Bearer {TOKEN}",
}


class HttpxTransport(Transport):
    """Custom Zeep transport using httpx.Client for synchronous HTTP requests."""

    def __init__(self) -> None:
        """Initialize the HttpxTransport with a synchronous httpx.Client."""
        super().__init__()
        self.client = httpx.Client()  # Synchronous HTTPX client

    def post(
        self,
        address: str,
        envelope: bytes | None,
        headers: dict,
    ) -> httpx.Response:
        """Override the default `post` method to use `httpx.Client`."""
        return self.client.post(
            address,
            data=envelope,  # type: ignore[reportOptionalMemberAccess]
            headers=headers,
        )


class _BaseClient:
    """Base class for common helper functions."""

    def __init__(self, version: APIVersion = DEFAULT_API_VERSION) -> None:
        """Initialize the base class."""
        _LOGGER.debug("Init API with version=%s", version.name)
        self._client: (
            httpx.Client | httpx.AsyncClient | SoapClient | AsyncSoapClient | None
        ) = None
        self._version = version
        self._username = None
        self._password = None
        self._user_id = None
        self._tanks = []

    def redact_response(self, obj: APIResponse) -> APIResponse:
        """Redact values email and password from request/response."""
        if isinstance(obj, dict):
            new_dict = {}
            for key, value in obj.items():
                if key in ("emailAddress", "password", "apiUserID", "signalmanNo"):
                    new_dict[key] = "*redacted*"
                else:
                    new_dict[key] = self.redact_response(value)
            return new_dict

        if isinstance(obj, list):
            return [self.redact_response(item) for item in obj]

        return obj

    def check_response(self, response: object) -> APIResponse:  # noqa: C901
        """Check the API response is well-formed or raise an exception."""

        def transform_soap_response(obj: APIResponse) -> APIResponse:
            """Convert a SOAP response to the same structure as the JSON API."""

            def lower_first(s: str) -> str:
                if s.startswith("API"):
                    return "api" + s[3:]
                return s[0].lower() + s[1:]

            if isinstance(obj, dict):
                result: dict = {}
                for k, v in obj.items():
                    if k in ("APILevel", "APITankInfo_V3", "APITankInfoItem"):
                        return transform_soap_response(v)
                    result[lower_first(k)] = transform_soap_response(v)
                return result

            if isinstance(obj, list):
                return [transform_soap_response(v) for v in obj]
            return obj

        if isinstance(response, httpx.Response):
            content = json.loads(response.content)
        else:
            content = transform_soap_response(serialize_object(response))

        if "apiResult" not in content or "code" not in content["apiResult"]:
            msg = "Malformed response from API: cannot extract response/code"
            _LOGGER.debug(
                "API error: %s, content=%s",
                msg,
                self.redact_response(content),
            )
            raise KingspanAPIError(msg)

        if content["apiResult"]["code"] != 0:
            msg = cast("str", content["apiResult"]["description"])
            _LOGGER.debug(
                "API error: %s, content=%s",
                msg,
                self.redact_response(content),
            )
            if "Authentication Failed" in msg:
                raise KingspanInvalidCredentialsError(msg)
            raise KingspanAPIError(content["apiResult"]["description"])

        _LOGGER.debug(
            "API request succeeded, content=%s",
            self.redact_response(content),
        )
        return content

    def handle_exception(self, exc: Exception) -> None:
        if isinstance(exc, httpx.TimeoutException):
            msg = f"HTTP request timeout: {exc}"
            _LOGGER.debug("API error: %s", msg)
            raise KingspanTimeoutError(msg) from exc

        msg = f"HTTP request failed: {exc}"
        _LOGGER.debug("API error: %s", msg)
        raise KingspanAPIError(msg) from exc

    def login_request(
        self,
        username: str,
        password: str,
    ) -> tuple[Callable, list, dict]:
        self._username = username
        self._password = password
        if self._version == APIVersion.CONNECT_V1:
            data = {"emailaddress": self._username, "password": self._password}
            client = cast("SoapClient | AsyncSoapClient", self._client)
            return (client.service.SoapMobileAPPAuthenicate_v3, [], data)

        data = {"emailAddress": self._username, "password": self._password}
        client = cast("httpx.Client | httpx.AsyncClient", self._client)
        return (
            client.post,
            [f"{API_BASE_URL}/v3/V3_SoapMobileApp/Authenticate_v3_Async"],
            {"headers": POST_HEADERS, "content": json.dumps(data)},
        )

    def login_response(
        self,
        response: httpx.Response | object,
        tank_type: type,
    ) -> None:
        if (
            hasattr(response, "status_code")
            and response.status_code == HTTP_UNAUTHORIZED
        ):
            # JSON API for an invalid JWT
            msg = "Invalid token authorization"
            raise KingspanAPIError(msg)
        data = self.check_response(response)
        self._user_id = data["apiUserID"]
        for tank_info in data["tanks"]:
            self._tanks.append(tank_type(self, tank_info["signalmanNo"]))

    def get_latest_level_request(
        self,
        signalman_no: str,
    ) -> tuple[Callable, list, dict]:
        data = {
            "userid": self._user_id,
            "password": self._password,
            "signalmanno": signalman_no,
            "culture": "en",
        }

        if self._version == APIVersion.CONNECT_V1:
            client = cast("SoapClient | AsyncSoapClient", self._client)
            return (client.service.SoapMobileAPPGetLatestLevel_v3, [], data)

        client = cast("httpx.Client | httpx.AsyncClient", self._client)
        return (
            client.post,
            [f"{API_BASE_URL}/v1/V1_SoapMobileApp/GetLatestLevel_v1_Async?culture=EN"],
            {"headers": POST_HEADERS, "content": json.dumps(data)},
        )

    def get_history_request(
        self,
        signalman_no: str,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> tuple[Callable, list, dict]:
        """Build a call history request for all clients."""
        start_date_dt = (
            datetime.fromtimestamp(0)  # noqa: DTZ006
            if start_date is None
            else start_date
        )
        end_date_dt = datetime.now() if end_date is None else end_date  # noqa: DTZ005

        if self._version == APIVersion.CONNECT_V1:
            data = {
                "userid": self._user_id,
                "password": self._password,
                "signalmanno": signalman_no,
                "startdate": start_date_dt.isoformat(),
                "enddate": end_date_dt.isoformat(),
            }
            client = cast("SoapClient | AsyncSoapClient", self._client)
            return (client.service.SoapMobileAPPGetCallHistory_v1, [], data)

        data = {
            "userIdPairWithSN": {
                "userId": self._user_id,
                "signalmanNo": signalman_no,
                "password": self._password,
            },
            "startDate": start_date_dt.isoformat(),
            "endDate": end_date_dt.isoformat(),
        }
        client = cast("httpx.Client | httpx.AsyncClient", self._client)
        return (
            client.post,
            [f"{API_BASE_URL}/v1/V1_SoapMobileApp/GetCallHistory_v1_Async"],
            {"headers": POST_HEADERS, "content": json.dumps(data)},
        )


class SensorClient(_BaseClient):
    """Synchronous client for interacting with the Kingspan Connect Sensor API."""

    def __init__(self, version: APIVersion = DEFAULT_API_VERSION) -> None:
        """Init for Synchronous client."""
        super().__init__(version)
        if self._version == APIVersion.CONNECT_V1:
            self._client = SoapClient(WSDL_URL, transport=HttpxTransport())
            self._client.set_ns_prefix(None, SOAP_NS_PREFIX)
        else:
            self._client = httpx.Client()

    def _request(self, func: Callable, *args: str | datetime | None) -> APIResponse:
        (func, req_args, req_kwargs) = func(*args)
        try:
            response = func(*req_args, **req_kwargs)
        except Exception as e:  # noqa: BLE001
            self.handle_exception(e)
        return response

    def login(self, username: str, password: str) -> None:
        """Authenticate with the API and initialize tank list."""
        response = self._request(self.login_request, username, password)
        self.login_response(response, Tank)

    @property
    def tanks(self) -> list[Tank]:
        """Return the list of Tank objects for the authenticated user."""
        return self._tanks

    def _get_latest_level(self, signalman_no: str) -> dict:
        """Tanks helper: get the latest level reading for a given tank."""
        response = self._request(self.get_latest_level_request, signalman_no)
        return self.check_response(response)

    def _get_history(
        self,
        signalman_no: str,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> list[dict]:
        """Tanks helper: get the history data for a given tank."""
        response = self._request(
            self.get_history_request,
            signalman_no,
            start_date,
            end_date,
        )
        new_response = self.check_response(response)
        return new_response["levels"]


class AsyncSensorClient(_BaseClient):
    """Asynchronous client for interacting with the Kingspan Connect Sensor API."""

    async def __aenter__(self) -> "AsyncSensorClient":
        """Enter the async context manager."""
        await self._init_client()
        return self

    async def __aexit__(self, exc_t, exc_v, exc_tb) -> None:  # noqa: ANN001
        """Exit the async context manager."""

    async def _init_client(self) -> None:
        # httpx.AsyncClient loads certificates in a blocking manner which is flagged
        # by Home Assistant, so we run in its own thread once on the first request.
        def _build_client() -> httpx.AsyncClient | AsyncSoapClient:
            if self._version == APIVersion.CONNECT_V1:
                wsdl_client = httpx.Client()
                httpx_client = httpx.AsyncClient()
                transport = AsyncTransport(client=httpx_client, wsdl_client=wsdl_client)
                client = AsyncSoapClient(WSDL_URL, transport=transport)
                client.set_ns_prefix(None, SOAP_NS_PREFIX)
                return client

            return httpx.AsyncClient()

        loop = get_running_loop()
        try:
            self._client = await loop.run_in_executor(None, _build_client)
        except Exception as e:  # noqa: BLE001
            self.handle_exception(e)

    async def _request(
        self,
        func: Callable,
        *args: str | datetime | None,
    ) -> APIResponse:
        """Private wrapper for SOAP/HTTP calls."""
        (func, req_args, req_kwargs) = func(*args)
        try:
            response = await func(*req_args, **req_kwargs)
        except Exception as e:  # noqa: BLE001
            self.handle_exception(e)
        return response

    async def login(self, username: str, password: str) -> None:
        """Authenticate with the API and initialize tank list."""
        response = await self._request(self.login_request, username, password)
        self.login_response(response, AsyncTank)

    @property
    async def tanks(self) -> list[AsyncTank]:
        """Return the list of Tank objects for the authenticated user."""
        return self._tanks

    async def _get_latest_level(self, signalman_no: str) -> dict:
        """Tanks helper: get the latest level reading for a given tank."""
        response = await self._request(self.get_latest_level_request, signalman_no)
        return self.check_response(response)

    async def _get_history(
        self,
        signalman_no: str,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> list[dict]:
        """Tanks helper: get the history data for a given tank."""
        response = await self._request(
            self.get_history_request,
            signalman_no,
            start_date,
            end_date,
        )
        new_response = self.check_response(response)
        return new_response["levels"]



def response(flow):
    def banner(s: str) -> None:
        print("-" * 50, s)

    server, port = flow.server_conn.address
    if "connectsensor.com" not in server:
        return

    print(f"SERVER: {server}:{port}")

    print(f"URL: {flow.request.url}")
    print(
        flow.request.method + " " + flow.request.path + " " + flow.request.http_version
    )

    banner("Request headers:")
    for k, v in flow.request.headers.items():
        print("%-20s: %s" % (k, v))
    banner("Request:")
    print(flow.request.content)

    banner("Response headers:")
    for k, v in flow.response.headers.items():
        print("%-20s: %s" % (k, v))
    banner("JSON response:")
    print(flow.response.content)
    banner("End of request")
